//  Copyright (C) SystemTools Software Inc. All Rights Reserved.
//
//
//  The STRCM utility is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
//  USA.
//
//  Further information on this software can be obtained from:
//
//  http://www.systemtools.com/strcm_frame.htm
//  
//  Additional contact and support information is available at this site.

// strcm.cpp : Defines the class behaviors for the application.
//

#include "stdafx.h"
#include "strcm.h"
#include "strcmDlg.h"

#include "stdio.h"
#include "winsvc.h"
#include "dialog.h"
#include "lmcons.h"
#include "lmshare.h"
#include "lmerr.h"
#include "lmapibuf.h"
#include "lmaccess.h"
#include "include.h"

#include "vncpwd.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CStrcmApp

BEGIN_MESSAGE_MAP(CStrcmApp, CWinApp)
	//{{AFX_MSG_MAP(CStrcmApp)
		// NOTE - the ClassWizard will add and remove mapping macros here.
		//    DO NOT EDIT what you see in these blocks of generated code!
	//}}AFX_MSG
	ON_COMMAND(ID_HELP, CWinApp::OnHelp)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CStrcmApp construction

CStrcmApp::CStrcmApp()
{
	// TODO: add construction code here,
	// Place all significant initialization in InitInstance
}

/////////////////////////////////////////////////////////////////////////////
// The one and only CStrcmApp object

CStrcmApp theApp;

/////////////////////////////////////////////////////////////////////////////
// CStrcmApp initialization

UINT CStrcmApp::GumbyThread(LPVOID param)
{
	GUMBYSTRUCT* gs = (GUMBYSTRUCT*) param;
	CStrcmDlg* m_dlg;
	m_dlg = (CStrcmDlg*) gs->m_App->m_pMainWnd; 
	
	// first, disable any way for the user to inadvertently close the app
	m_dlg->m_cOK.EnableWindow(FALSE);
	m_dlg->m_cCancel.EnableWindow(FALSE);

	// now, wait for the process to terminate
	WaitForSingleObjectEx(gs->m_hProcess, INFINITE, FALSE);

	// report that it closed
	gs->m_App->DisplayProgressMsg(gs->m_wnd, _T("   ...Viewer has closed."));

	// then, we close the handles
	CloseHandle(gs->m_hProcess);
	CloseHandle(gs->m_hThread);

	// stop the service and do any other options
	CString StopServiceName = gs->m_App->StopServices(gs->m_ServiceName, gs->m_Host, gs->m_wnd);

	if (!gs->m_UnInstallService)
		goto bailout;
	
	// at this point, we've stopped the service and now need to uninstall it
	if (StopServiceName.GetLength() == 0)
	{
		CString temp;
		temp.LoadString(IDS_NOSERVICE_TO_UNINSTALL);
		gs->m_App->ReportError(temp);
		goto bailout;
	}

	gs->m_App->RemoteUnInstall(gs->m_ConfigFile, gs->m_Host, gs->m_wnd);	
    goto bailout;

bailout:
	gs->m_App->DisplayProgressMsg(gs->m_wnd, _T(" "));
	m_dlg->m_cOK.EnableWindow(TRUE);
	m_dlg->m_cCancel.EnableWindow(TRUE);
	if (m_dlg->m_AutoTerminateAfterViewing)
		m_dlg->Exit();
	return 1;
}

BOOL CStrcmApp::InitInstance()
{
	Enable3dControls();			// Call this when using MFC in a shared DLL

	CStrcmDlg dlg;
	m_pMainWnd = &dlg;
	CString CmdLine, Cmd, temp, FileName, HostName, Operation;
	int i;
	BOOL Ready = FALSE;

	// yes, there are better ways to do this, but lets just do it..
	CmdLine = m_lpCmdLine;
	if (CmdLine.GetLength() > 0)
	{
		i = 0;
		do
		{
			i = CmdLine.ReverseFind('/');	
			if (i != -1)
			{
				temp = CmdLine.Mid(i+1);
				CmdLine = CmdLine.Left(i-1);
				// now figure out what our cmd is
				i = temp.Find('=');
				if (i != -1)
				{
					Cmd = temp.Left(i);
					temp = temp.Mid(i+1);
					if (Cmd.CompareNoCase(_T("rcm")) == 0)
						FileName = temp;
					else if (Cmd.CompareNoCase(_T("host")) == 0)
						HostName = temp;
					else if (Cmd.CompareNoCase(_T("cmd")) == 0)
						Operation = temp;
				}
			}
		}
		while (i != -1);
	}

	// if there isn't any path to the file, we need to set the working directory to be
	// the current directory
	if (FileName.GetLength() > 0)
	{
		if (FileName.Find('\\') == -1)
			FileName = GetCurrentDir() + FileName;
		if ((HostName.GetLength() > 0) && (Operation.GetLength() > 0))
			Ready = TRUE;
	}

	dlg.m_AutoRun = FALSE;
	dlg.m_AutoTerminateAfterViewing = FALSE;
	if (Ready)
	{
		temp = GetSetting(FileName, _T("General"), _T("AutoExecute"));
		if (temp == '1')
			dlg.m_AutoRun = TRUE;
		temp = GetSetting(FileName, _T("View"), _T("AutoTerminateAfterViewing"));
		if (temp == '1')
			dlg.m_AutoTerminateAfterViewing = TRUE;
	}
	dlg.m_FileName = FileName;
	dlg.m_Host = HostName;
	dlg.m_App = this;
	if (Operation.CompareNoCase(_T("view")) == 0)
		dlg.m_Operation = 0;
	else if (Operation.CompareNoCase(_T("install")) == 0)
		dlg.m_Operation = 1;
	else if (Operation.CompareNoCase(_T("uninstall")) == 0)
		dlg.m_Operation = 2;
	else if (Operation.CompareNoCase(_T("options")) == 0)
		dlg.m_Operation = 3;
	else if (Operation.CompareNoCase(_T("start")) == 0)
		dlg.m_Operation = 4;
	else if (Operation.CompareNoCase(_T("stop")) == 0)
		dlg.m_Operation = 5;
	int nResponse = dlg.DoModal();

	//if (nResponse == IDOK)
	//else if (nResponse == IDCANCEL)

	// Since the dialog has been closed, return FALSE so that we exit the
	//  application, rather than start the application's message pump.
	return FALSE;
}

void CStrcmApp::RemoteViewer(LPCTSTR CmdFile, LPCTSTR Host, CWnd* wnd)
{
	CString temp;
	DWORD Err;

	BOOL bCheckService = FALSE;
	temp = GetSetting(CmdFile, _T("View"), _T("AutoInstallBeforeViewing"));
	if (temp == '1')
		bCheckService = TRUE;
	if (bCheckService)
	{
		DisplayProgressMsg(wnd, _T("Checking remote software installation..."));
		CString Dest;

		temp = GetSetting(CmdFile, _T("Install"), _T("ServiceExeName"));
		if (temp.GetLength() == 0)
		{
			temp.LoadString(IDS_NO_SERVICE_EXENAME);
			ReportError(temp);
			return;
		}
		Dest = GetSetting(CmdFile, _T("Install"), _T("DestinationDirectory"));
		if (Dest.GetLength() == 0)
		{
			temp.LoadString(IDS_NO_SERVICE_DESTDIR);
			ReportError(temp);
			return;
		}
		Dest.Replace(_T("%computer%"), Host);
		if (Dest.Right(1) != '\\')
			Dest += '\\';
		Dest += temp;

		if (!CheckIfExists(Dest))
		{
			if (!RemoteInstall(CmdFile, Host, wnd))
			{
				if (AfxMessageBox(IDS_INSTALLERROR_CONTINUE, MB_YESNO | MB_DEFBUTTON2) != IDYES)
				{
					return;
					DisplayProgressMsg(wnd, _T("Installation aborted..."));
					DisplayProgressMsg(wnd, _T(" "));
				}
			}
		}
	}

	CString Cmd, StartServiceName, StopServiceName, ServiceNameList;
	BOOL bStopService = FALSE;
	BOOL bUnInstallService = FALSE;
	BOOL bStartService = FALSE;

	temp = GetSetting(CmdFile, _T("View"), _T("SetOptionsBeforeViewing"));
	if (temp == '1')
		SetVncOptions(CmdFile, Host, _T(""), FALSE);

	ServiceNameList = GetSetting(CmdFile, _T("Install"), _T("ServiceNames"));
	temp = GetSetting(CmdFile, _T("View"), _T("StartServiceBeforeViewing"));
	if (temp == '1')
		bStartService = TRUE;
	if (bStartService)
	{
		if (ServiceNameList.GetLength() > 0)
			StartServiceName = StartServices(ServiceNameList, Host, wnd);
	}

	temp = GetSetting(CmdFile, _T("View"), _T("StopServiceAfterViewing"));
	if (temp == '1')
		bStopService = TRUE;
	temp = GetSetting(CmdFile, _T("View"), _T("UnInstallAfterViewing"));
	if (temp == '1')
	{
		bUnInstallService = TRUE;
		bStopService = TRUE;
	}

	Cmd = GetSetting(CmdFile, _T("View"), _T("ViewerCommand"));
	Cmd.Replace(_T("%computer%"), Host);
	if (Cmd.Find('\\') == -1)
	{
		temp = GetCurrentDir();
		if (temp.GetLength() > 0)
			Cmd = temp + Cmd;
	}
		
	STARTUPINFO si;
	PROCESS_INFORMATION pi;

	ZeroMemory(&si, sizeof(si));
	si.cb = sizeof(si);
	si.dwFlags = STARTF_FORCEONFEEDBACK; 
	si.wShowWindow = SW_SHOW;

	BOOL BOOLres = CreateProcess(NULL, (LPTSTR) ((LPCTSTR) Cmd), NULL, NULL, FALSE, 
			          NORMAL_PRIORITY_CLASS, NULL, NULL, &si, &pi);
	if (BOOLres == FALSE)
	{
		Err = GetLastError();
		temp.Format(IDS_EXECUTE_ERROR, Cmd);
		temp += LookupError(Err);
		ReportError(temp);
	}
	else
	{
		if (bStopService)
		{
			DisplayProgressMsg(wnd, _T("Waiting for viewer to close..."));

			// Note: while the line below works, it prevents the application from 
			// getting refresh/paint messages, so the GUI becomes hosed quickly.
			//WaitForSingleObjectEx(pi.hProcess, INFINITE, FALSE);

			GUMBYSTRUCT* gs = new GUMBYSTRUCT;
			gs->m_App = this;
			//gs->m_AppDlg = dlg;
			gs->m_hProcess = pi.hProcess;
			gs->m_hThread = pi.hThread;
			gs->m_wnd = wnd;
			if (StartServiceName.GetLength() > 0)
				gs->m_ServiceName = StartServiceName;
			else
				gs->m_ServiceName = ServiceNameList;
			gs->m_ConfigFile = CmdFile;
			gs->m_Host = Host;
			gs->m_UnInstallService = bUnInstallService;
			AfxBeginThread(GumbyThread, gs);
		}
		else
		{
			CloseHandle(pi.hProcess);
			CloseHandle(pi.hThread);
			// if we are not going to wait around, see if we need to exit
			CStrcmDlg* m_dlg;
			m_dlg = (CStrcmDlg*) m_pMainWnd; 
			if (m_dlg->m_AutoTerminateAfterViewing)
				m_dlg->Exit();
		}
	}
}

CString CStrcmApp::GetSetting(LPCTSTR CmdFile, LPCTSTR Section, LPCTSTR Setting)
{
	TCHAR Value[256];

	GetPrivateProfileString(Section, Setting, _T(""), Value, sizeof(Value), CmdFile);
	return Value;
}

int CStrcmApp::GetSettingInt(LPCTSTR CmdFile, LPCTSTR Section, LPCTSTR Setting, int Default)
{
	return GetPrivateProfileInt(Section, Setting, Default, CmdFile);
}

void CStrcmApp::SetSettingInt(LPCTSTR CmdFile, LPCTSTR Section, LPCTSTR Setting, int Value)
{
	CString temp;
	temp.Format(_T("%d"), Value);
	WritePrivateProfileString(Section, Setting, temp, CmdFile);
}

void CStrcmApp::ReportError(LPCTSTR ErrorMsg)
{
	MessageBeep(MB_ICONSTOP);
	AfxMessageBox(ErrorMsg, MB_ICONSTOP);
}

CString CStrcmApp::LookupError(DWORD Err)
{
	LPVOID lpMsgBuf;
	DWORD dwErr;
	HMODULE hModule = NULL; // default to system source
 
    //
    // if dwLastError is in the network range, load the message source
    //
	if (Err >= NERR_BASE && Err <= MAX_NERR) 
		hModule = LoadLibraryEx(TEXT("netmsg.dll"), NULL, LOAD_LIBRARY_AS_DATAFILE);

	dwErr = FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER |
        FORMAT_MESSAGE_IGNORE_INSERTS |
        FORMAT_MESSAGE_FROM_SYSTEM | // always consider system table
        ((hModule != NULL) ? FORMAT_MESSAGE_FROM_HMODULE : 0),
        hModule, // module to get message from (NULL == system)
        Err,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // default language
        (LPTSTR) &lpMsgBuf,
        0,
        NULL);

	if(hModule != NULL)
      FreeLibrary(hModule);

	if (dwErr == 0)
		return "";

	CString Msg((LPCTSTR) lpMsgBuf);
	Msg = " " + Msg;
	// Free the buffer.
	LocalFree( lpMsgBuf );
	return Msg;
}

CString CStrcmApp::GetCurrentDir()
{
	CString temp;
	CWinApp* myApp = AfxGetApp();
	int i;

	temp = myApp->m_pszHelpFilePath;
	i = temp.ReverseFind('\\');
	if (i > 0)
		temp = temp.Left(i);
	else
	{
		TCHAR CurrentDir[MAX_PATH];
		DWORD sz = sizeof(CurrentDir) / sizeof(TCHAR);
		GetCurrentDirectory(sz, CurrentDir);
		temp = CurrentDir;
	}

	if ((temp.GetLength() > 0) && (temp.Right(1) != '\\'))
		temp += '\\';
	return temp;
}

BOOL CStrcmApp::CheckIfExists(LPCTSTR File)
{
	HANDLE hFind;
	WIN32_FIND_DATA fd;
		
	hFind = ::FindFirstFile(File, &fd);
	if (hFind == INVALID_HANDLE_VALUE)
		return FALSE;
	else
	{
		::FindClose(hFind);
		return TRUE;
	}
}

BOOL CStrcmApp::GetForceX64Setting(LPCTSTR CmdFile)
{
	CString temp = GetSetting(CmdFile, _T("General"), _T("ForceX64Registry"));
	if (temp == '1')
		return TRUE;
	else
		return FALSE;
}

REGSAM CStrcmApp::GetRegAccessRights(REGSAM Default, BOOL ForceX64)
{
	if (ForceX64)
		Default |= KEY_WOW64_64KEY;
	return Default;
}

BOOL CStrcmApp::RemoteInstall(LPCTSTR CmdFile, LPCTSTR Host, CWnd* wnd)
{
	CString temp = GetSetting(CmdFile, _T("Install"), _T("EULA"));
	if (temp.GetLength() > 0)
	{
		if (temp.Find('\\') == -1) // file name only given...append current directory
			temp = GetCurrentDir() + temp;
		CFile cfile;
		if (cfile.Open(temp, CFile::modeRead))
		{
			DWORD Len = cfile.GetLength();
			LPTSTR szLine = (LPTSTR) HeapAlloc(GetProcessHeap(), 0, Len);
			UINT nBytesRead = cfile.Read(szLine, Len);
			cfile.Close();

			CEula EULA;
			EULA.m_Accept = TRUE;
			EULA.m_Eula = szLine;
			HeapFree(GetProcessHeap(), 0, szLine);
			EULA.DoModal();
			if (!EULA.m_Accept)
				return FALSE;
		}
		else
		{
			CString Error;
			Error.Format(IDS_EULA_NOTFOUND, temp);
			ReportError(Error);
			return FALSE;
		}
	}

	BOOL ret = FALSE;
	BOOL InstallService = FALSE;
	BOOL CopyErrorsFound = FALSE;
	BOOL InstallVNCServiceHelper = FALSE;
	CString SourceDir, DestDir, FileList, ServiceName, ServiceDisplayName, ServiceExeName;
	CString SourceFile, TargetFile, SoftwareType, ServiceCmd, LocalDestPath, ShareName, Password;
	DWORD StartType, Err;
	int i;
	HKEY hKey = NULL;
	HKEY hKey2 = NULL;
	SC_HANDLE sch = NULL;
	SC_HANDLE sch2 = NULL;
	BOOL ForceX64 = FALSE;

	temp = GetSetting(CmdFile, _T("Install"), _T("ServiceAutoStart"));
	if (temp == '1')
		StartType = SERVICE_AUTO_START;
	else
		StartType = SERVICE_DEMAND_START;
	temp = GetSetting(CmdFile, _T("Install"), _T("InstallService"));
	if (temp == '1')
		InstallService = TRUE;
	SourceDir = GetSetting(CmdFile, _T("Install"), _T("SourceDirectory"));
	DestDir = GetSetting(CmdFile, _T("Install"), _T("DestinationDirectory"));
	FileList = GetSetting(CmdFile, _T("Install"), _T("FileList"));
	ServiceName = GetSetting(CmdFile, _T("Install"), _T("ServiceNames"));
	ServiceDisplayName = GetSetting(CmdFile, _T("Install"), _T("ServiceDisplayName"));
	ServiceExeName = GetSetting(CmdFile, _T("Install"), _T("ServiceExeName"));
	SoftwareType = GetSetting(CmdFile, _T("General"), _T("SoftwareType"));
	Password = "";
	ForceX64 = GetForceX64Setting(CmdFile);

	// prompt for parameters (override and/or confirm)
	// in case we want to make this an option (ie, a silent install), we do it this way...
	CInstall install;
	install.m_Source = SourceDir;
	install.m_Type = SoftwareType;
	install.m_Target.Format(IDS_INSTALL_TARGET_MSG, Host);
	install.m_Dest = DestDir;
	install.m_InstallService = InstallService;
	install.m_DisplayName = ServiceDisplayName;
	install.m_ExeName = ServiceExeName;
	install.m_ServiceName = ServiceName;
	if (StartType == SERVICE_AUTO_START)
		install.m_AutoStart = TRUE;
	if (install.DoModal() != IDOK)
		return FALSE;

	UpdateWindow(m_pMainWnd->m_hWnd); // add this in order to update window for recording

	// now put the parameters back...
	SourceDir = install.m_Source;
	DestDir = install.m_Dest;
	InstallService = install.m_InstallService;
	if (install.m_AutoStart)
		StartType = SERVICE_AUTO_START;
	else
		StartType = SERVICE_DEMAND_START;
	ServiceDisplayName = install.m_DisplayName;
	ServiceExeName = install.m_ExeName;
	ServiceName = install.m_ServiceName;
	Password = install.m_Password;
	
	// now fix up settings
	if (DestDir.GetLength() > 0)
	{
		DestDir.Replace(_T("%computer%"), Host);
		// create the destination directory
		CreateDirectory(DestDir, NULL); // we ignore errors here
		if (DestDir.Right(1) != '\\')
			DestDir += '\\';
	}
	if (SourceDir.GetLength() == 0)
		SourceDir = GetCurrentDir();
	else
	{
		// see if we have just a name, which would be just a directory name
		if (SourceDir.Find('\\') == -1) // no directory entry
			SourceDir = GetCurrentDir() + SourceDir;
		if (SourceDir.Right(1) != '\\') // make sure we end in a slash
			SourceDir += '\\';
	}

	// copy the files
	if ((SourceDir.GetLength() > 0) && (DestDir.GetLength() > 0) &&	(FileList.GetLength() > 0))
	{
		DisplayProgressMsg(wnd, _T("Copying files..."));
		if (FileList.Right(1) != ';')
			FileList += ';';
		SourceFile = "";
		for (i=0; i < FileList.GetLength(); i++)
		{
			if (FileList[i] != ';')
				SourceFile += FileList[i];
			else
			{
				TargetFile = DestDir + SourceFile;
				SourceFile = SourceDir + SourceFile;
				if (!CopyFile(SourceFile, TargetFile, FALSE))
				{
					Err = GetLastError();
					temp.Format(IDS_COPYFILE_ERROR, SourceFile, TargetFile);
					temp += LookupError(Err);
					ReportError(temp);
					CopyErrorsFound = TRUE;
				}
				SourceFile = "";
			}
		}
	}

	// see if we got errors copying files and ask if we should continue
	if (CopyErrorsFound)
	{
		if (AfxMessageBox(IDS_COPYERROR_CONTINUE, MB_YESNO | MB_DEFBUTTON2) != IDYES)
		{
			DisplayProgressMsg(wnd, _T("Installation aborted..."));
			DisplayProgressMsg(wnd, _T(" "));
			return FALSE;
		}
	}


	// check to see if we should do the VNC Helper settings
	temp = GetSetting(CmdFile, _T("VNC Options"), _T("InstallServiceHelper"));
	if (temp == '1')
		InstallVNCServiceHelper = TRUE;

	// now comes a tricky part.  We need to fix up ServiceExeName for local drive letters
	// The remote service .exe file must specify a local path on the remote computer.  
	// To get a local path, we need to do some conversions
	// This should be put into a separate function...we only need to do this if we are 
	// going to install the service or helper feature for VNC

	if ((InstallService) || (InstallVNCServiceHelper))
	{
		// first, we get the share name that is part of the destination directory
		i = DestDir.Find(_T('\\'), 2);
		if (i == -1)
		{
			temp.Format(IDS_PARSE_ERROR, DestDir);
			ReportError(temp);
			return FALSE;
		}
		temp = DestDir.Mid(i + 1);
		i = temp.Find(_T('\\'), 0);
		if (i == -1) // we only have a unc and share
		{
			ShareName = temp;
			temp = "";
		}
		else
		{
			ShareName = temp.Left(i);
			temp = temp.Mid(i + 1);
		}
		// NOTE: temp now holds any extra directory on our share
	
		PSHARE_INFO_2 si = NULL;
		NET_API_STATUS NetErr;
		WCHAR wShareName[256];
		WCHAR wHost[256];

		// because this is not a UNICODE program, we have to convert 
		// ShareName and Host to Unicode
		mbstowcs(wShareName, ShareName, 256);
		mbstowcs(wHost, Host, 256);
	
		NetErr = NetShareGetInfo((LPWSTR) wHost, (LPWSTR) wShareName, 2, (LPBYTE*) &si);
		if (NetErr != NERR_Success)
		{
			temp.Format(IDS_SHARE_LOOKUP_ERROR, ShareName, Host);
			temp += LookupError(NetErr);
			ReportError(temp);
			return FALSE;
		}
		// ...and because this is not a UNICODE program, we have to convert the results
		// back into something that we can use.
		char Path[256];
		wcstombs(Path, (LPWSTR) si->shi2_path, 256);

		// and now, FINALLY, we can put it all back together
		LocalDestPath = Path;
		if (LocalDestPath.Right(1) != '\\')
			LocalDestPath += '\\';
		if (temp.GetLength() > 0) //temp contains our extra path
		{
			LocalDestPath += temp;
			if (LocalDestPath.Right(1) != '\\')
				LocalDestPath += '\\';
		}
		// now we free our buffer (to prevent a memory leak)
		NetApiBufferFree(si);
	}

	// set the password and other options depending upon the type of software
	// this generally should/must be done prior to installing the service

	if (SoftwareType.CompareNoCase(_T("VNC")) == 0)
	{
		SetVncOptions(CmdFile, Host, Password, FALSE);
		
		// check to see if we should install the VNC service helper reg entry
		if (InstallVNCServiceHelper)
		{
			DisplayProgressMsg(wnd, _T("Installing VNC Service Helper..."));
			if (!RegConnectRegistry(Host, HKEY_LOCAL_MACHINE, &hKey) == ERROR_SUCCESS)
			{
				Err = GetLastError();
				temp.Format(IDS_CONNECT_REGISTRY_ERROR3, Host);
				temp += LookupError(Err);
				ReportError(temp);
				goto bailout;
			}
			if (RegCreateKeyEx(hKey, _T("Software\\Microsoft\\Windows\\CurrentVersion\\Run"), 0, NULL, 0, GetRegAccessRights(KEY_SET_VALUE, ForceX64), NULL, &hKey2, NULL) != ERROR_SUCCESS)
			{
				Err = GetLastError();
				temp.Format(IDS_VNC_SERVICE_HELPER_ERROR, Host);
				temp += LookupError(Err);
				ReportError(temp);
				goto bailout;
			}

			ServiceCmd = _T("\"") + LocalDestPath + ServiceExeName + _T("\"") + _T(" -servicehelper");
			if (RegSetValueEx(hKey2, _T("WinVNC"), 0, REG_SZ, (CONST BYTE*) (LPCTSTR) ServiceCmd, ServiceCmd.GetLength() + 1) != ERROR_SUCCESS)
			{
				Err = GetLastError();
				temp.Format(IDS_VNC_SERVICE_HELPER_ERROR, Host);
				temp += LookupError(Err);
				ReportError(temp);
				goto bailout;
			}
			RegFlushKey(hKey2);
			RegFlushKey(hKey);
		}
	}

	if (InstallService)
	{
		// create the service
		DisplayProgressMsg(wnd, _T("Installing service..."));
		ServiceCmd = _T("\"") + LocalDestPath + ServiceExeName + _T("\"");
		if (SoftwareType.CompareNoCase(_T("VNC")) == 0)
			ServiceCmd += _T(" -service");

		sch = OpenSCManager(Host, SERVICES_ACTIVE_DATABASE, SC_MANAGER_CREATE_SERVICE);
		if (sch == NULL)
		{
			Err = GetLastError();
			temp.LoadString(IDS_OPEN_SERVICE_DB_ERROR);
			temp += LookupError(Err);
			ReportError(temp);
			goto bailout;
		}
		sch2 = CreateService(
			sch, 
			ServiceName, 
			ServiceDisplayName,
			SERVICE_ALL_ACCESS,
			SERVICE_WIN32_OWN_PROCESS | SERVICE_INTERACTIVE_PROCESS,
			StartType,
			SERVICE_ERROR_NORMAL,
			ServiceCmd,
			NULL,
			NULL, //tag
			NULL, //dependencies
			NULL, //account name
			NULL); //password

		if (sch2 == NULL)
		{
			Err = GetLastError();
			temp.Format(IDS_CREATE_SERVICE_ERROR, Host);
			temp += LookupError(Err);
			ReportError(temp);
			goto bailout;
		}
	}

	ret = TRUE; // we assume that this is all that we absolutely have to do

bailout:
	if (sch)
		CloseServiceHandle(sch);
	if (sch2)
		CloseServiceHandle(sch2);
	if (hKey)
		RegCloseKey(hKey);
	if (hKey2)
		RegCloseKey(hKey2);
	temp.Format(_T("Installation complete on %s."), Host);
	DisplayProgressMsg(wnd, temp);
	return ret;
}

void CStrcmApp::RemoteUnInstall(LPCTSTR CmdFile, LPCTSTR Host, CWnd* wnd)
{
	CString temp;
	DWORD Err;
	DWORD SleepTime = 2000;
	DWORD SleepRetryLimit = 3;
	BOOL Try;
	int i, SleepLoop;

	// determine software type
	CString SoftwareType = GetSetting(CmdFile, _T("General"), _T("SoftwareType"));
	if (SoftwareType.GetLength() == 0)
	{
		temp.LoadString(IDS_NO_SOFTWARE_TYPE);
		ReportError(temp);
		return;
	}
	
	// first, we should stop the service
	CString ServiceNameList, StopServiceName;
	ServiceNameList = GetSetting(CmdFile, _T("Install"), _T("ServiceNames"));
	if (ServiceNameList.GetLength() > 0)
		StopServiceName = StopServices(ServiceNameList, Host, wnd);
	if (StopServiceName.GetLength() > 0)
	{
		SC_HANDLE sch = OpenSCManager(Host, SERVICES_ACTIVE_DATABASE, SC_MANAGER_ALL_ACCESS);
		if (sch == NULL)
		{
			Err = GetLastError();
			temp.LoadString(IDS_OPEN_SERVICE_DB_ERROR2);
			temp += LookupError(Err);
			ReportError(temp);
		}
		else
		{
			SC_HANDLE sch2 = OpenService(sch, StopServiceName, DELETE);
			if (sch2 == NULL)
			{
				Err = GetLastError();
				temp.LoadString(IDS_OPEN_SERVICE_ERROR2);
				temp += LookupError(Err);
				ReportError(temp);
				CloseServiceHandle(sch);
				return;
			}
			else
			{
				if (!DeleteService(sch2))
				{
					Err = GetLastError();
					temp.LoadString(IDS_DELETE_SERVICE_ERROR);
					temp += LookupError(Err);
					ReportError(temp);
				}
				CloseServiceHandle(sch2);
			}
			CloseServiceHandle(sch);
		}
	}

	// due to a quirk in how some services 'stop', we need to wait around a little while
	// before nuking everything, otherwise we may get file lock, etc. errors.
	DisplayProgressMsg(wnd, _T("Waiting..."));
	Sleep(5000);

	// then, we can remove registry entries (specific to each software type)
	HKEY hKey;
	DisplayProgressMsg(wnd, _T("Deleting registry entries..."));
	if (!RegConnectRegistry(Host, HKEY_LOCAL_MACHINE, &hKey) == ERROR_SUCCESS)
	{
		Err = GetLastError();
		temp.Format(IDS_CONNECT_REGISTRY_ERROR2, Host);
		temp += LookupError(Err);
		ReportError(temp);
	}
	else
	{
		if (SoftwareType.CompareNoCase(_T("VNC")) == 0)
			DeleteVNCRegistry(CmdFile, hKey, Host);
		RegCloseKey(hKey);
	}

	// then, remove files

	CString DestDir = GetSetting(CmdFile, _T("Install"), _T("DestinationDirectory"));
	CString FileList = GetSetting(CmdFile, _T("Install"), _T("FileList"));
	if ((DestDir.GetLength() == 0) || (FileList.GetLength() == 0))
	{
		temp.LoadString(IDS_NO_FILES_TO_DELETE_ERR);
		ReportError(temp);
		return;
	}
	// now fix up settings
	
	DestDir.Replace(_T("%computer%"), Host);
	if (DestDir.Right(1) != '\\')
		DestDir += '\\';
	
	// delete the files
	if (FileList.Right(1) != ';')
		FileList += ';';
	CString SourceFile = "";

	DisplayProgressMsg(wnd, _T("Deleting files..."));
	for (i=0; i < FileList.GetLength(); i++)
	{
		if (FileList[i] != ';')
			SourceFile += FileList[i];
		else
		{
			SourceFile = DestDir + SourceFile;
			Try = TRUE;
			SleepLoop = 0;
			do
			{
				if (!DeleteFile(SourceFile))
				{
					Err = GetLastError();
					if (Err == ERROR_FILE_NOT_FOUND)
						Try = FALSE; // get out, as the file is not there anyway
					else
					{
						SleepLoop++;
						if (SleepLoop <= SleepRetryLimit)
							Sleep(SleepTime * SleepLoop); // sleep then try again
						else
						{
							temp.Format(IDS_DELETEFILE_ERROR, SourceFile);
							temp += LookupError(Err);
							ReportError(temp);
							Try = FALSE; // set flag to get out, as we have tried enough now
						}
					}
				}
				else
					Try = FALSE; // file was deleted (ie no error)
			}
			while (Try);
			SourceFile = "";
		}
	}

	// finally delete the directory
	if (DestDir.Right(1) == '\\')
		DestDir = DestDir.Left(DestDir.GetLength() - 1);
	if (!RemoveDirectory(DestDir))
	{
		Err = GetLastError();
		if (Err != ERROR_FILE_NOT_FOUND)
		{
			temp.Format(IDS_DELETEDIR_ERROR, DestDir);
			temp += LookupError(Err);
			ReportError(temp);
		}
	}
	temp.Format(_T("UnInstallion complete on %s."), Host);
	DisplayProgressMsg(wnd, temp);
}
	

void CStrcmApp::DeleteVNCRegistry(LPCTSTR CmdFile, HKEY hKey, LPCTSTR Host)
{
	HKEY hKey2 = NULL;
	CString temp, Setting, RegKey, SpecialKey;
	DWORD Err, BigBlobSize, Values;
	HKEY hKeySpecial = NULL;
	LPTSTR BigBlob = NULL;
	BOOL FoundSpecial, DefaultInstallationKey, ForceX64;
	int i, j, f, Levels;
	CStringArray RegistryPaths;
	LONG ret;

	ForceX64 = GetForceX64Setting(CmdFile);

	// new in v1.13...only remove this setting if it was configured to be installed
	temp = GetSetting(CmdFile, _T("VNC Options"), _T("InstallServiceHelper"));
	if (temp == '1')
	{
		if (RegOpenKeyEx(hKey, _T("Software\\Microsoft\\Windows\\CurrentVersion\\Run"), 0, GetRegAccessRights(KEY_SET_VALUE, ForceX64), &hKey2) == ERROR_SUCCESS)
		{
			if (RegDeleteValue(hKey2, _T("WinVNC")) != ERROR_SUCCESS)
			{
				Err = GetLastError();
				temp.Format(IDS_REG_DELETE_KEY, Host);
				temp += LookupError(Err);
				ReportError(temp);
			}
			RegCloseKey(hKey2);
			hKey2 = NULL;
		}
	}
	// end new

	CString RegBase = GetSetting(CmdFile, _T("Install"), _T("RegistryKeyLocation"));
	if (RegBase.GetLength() == 0)
		RegBase = "SOFTWARE\\ORL\\WinVNC3\\Default";

	// new in v1.13
	// In order to delete any custom entries (values) that were added, we need to look for any custom entries, and delete them one-by-one.
	// If not specified, we will delete them from the default key location, or the key specified above for RegistryKeyLocation
	// Once done deleting all of the values, we will then go back and clean up the registry key paths from our table

	// first, store the default registry key location into our table
	RegistryPaths.Add(RegBase);

	// open our default key, as we will probably be deleting values from it.   If we get an error, its unclear what we should do, as the 
	// entry may already be deleted, which is OK, since we are going to have to delete it anyway, or the configuration is to use some alternate
	// key paths, and we will take care of those later.
	if (RegOpenKeyEx(hKey, RegBase, 0, GetRegAccessRights(KEY_SET_VALUE, ForceX64), &hKey2) != ERROR_SUCCESS)
		hKey2 = NULL; 
	
	// allocate some memory for storage of all of our values
	BigBlob = (LPTSTR) HeapAlloc(GetProcessHeap(), 0, BIG_BLOB_SIZE); //more than enough room
	if (BigBlob == NULL)
	{
		// TODO: display error...should never happen
		return;
	}

	BigBlobSize = GetPrivateProfileSection(_T("VNC Settings"), BigBlob, BIG_BLOB_SIZE, CmdFile);

	// now go through our buffer of all of the settings, and delete each value
	Setting = "";
	for (i=0; i < BigBlobSize; i++)
	{
		if (BigBlob[i] != '\0')
		{
			Setting += BigBlob[i];
			continue;
		}
		
		// look to see if the default location has been overridden
		FoundSpecial = FALSE;
		f = Setting.ReverseFind('\\');
		if (f != -1)
		{
			// dang, now we have to do more work
			FoundSpecial = TRUE;
			SpecialKey = Setting.Left(f);
			RegistryPaths.Add(SpecialKey); // add it to our table
			Setting = Setting.Mid(f + 1);
			// open our special key
			ret = RegOpenKeyEx(hKey, SpecialKey, 0, GetRegAccessRights(KEY_SET_VALUE, ForceX64), &hKeySpecial);
			if (ret != ERROR_SUCCESS)
			{
				temp.Format(IDS_OPEN_REGISTRY_KEY_ERROR, SpecialKey);
				temp += LookupError(ret);
				ReportError(temp);
				Setting = "";
				continue;
			}
		}

		// look for the '=' and parse out the settings vs. the value
		f = Setting.Find('=');
		if (f == -1) // something is wrong with the format
		{
			Setting = "";
			continue;
		}
		
		Setting = Setting.Left(f);  // we discard the value portion as we dont care; we are going to delete it

		ret = 0;
		if (hKeySpecial)
		{
			ret = RegDeleteValue(hKeySpecial, Setting);
			RegCloseKey(hKeySpecial);
			hKeySpecial = NULL;
		}
		else if (hKey2)
			ret = RegDeleteValue(hKey2, Setting);
		if (ret != ERROR_SUCCESS)
		{
			// again, what can we do ?  It may be normal that the key is not there...
		}

		Setting = ""; // clear our setting for the next one

	}

	if (BigBlob)
		HeapFree(GetProcessHeap(), 0, BigBlob);
	if (hKey2)
		RegCloseKey(hKey2);

	// now we need to go through each of our key paths and delete them, but only if they are empty of values
	for (i=0; i < RegistryPaths.GetSize(); i++)
	{
		SpecialKey = RegistryPaths.GetAt(i);
		if (SpecialKey == RegBase)
			DefaultInstallationKey = TRUE;
		else
			DefaultInstallationKey = FALSE;

		// we need to count the number of levels in our key path (we don't want to delete the \software key!)
		Levels = 0;
		for (j=0; j < SpecialKey.GetLength(); j++)
		{
			if (SpecialKey[j] == '\\')
				Levels++;
		}
	
		// now go through each level in the registry path and process it
		for (j=0; j < Levels; j++)
		{
			f = SpecialKey.ReverseFind('\\');
			RegKey = SpecialKey.Mid(f+1);
			SpecialKey = SpecialKey.Left(f);

			// just in case...
			if (RegKey.CompareNoCase(_T("Software")) == 0)
				break;

			// now we need to see if our key has anything in it.  If it does, then either our deletions failed, or there was something there from 
			// another application, etc. and we probably should not delete the key
			// EXCEPTION: we always delete the entire contents of our default installation key, as the password value (and possibly others)
			// will be populated upon usage of the software
			
			if (RegOpenKeyEx(hKey, SpecialKey, 0, GetRegAccessRights(KEY_SET_VALUE | KEY_QUERY_VALUE, ForceX64), &hKey2) == ERROR_SUCCESS)
			{
				if (!DefaultInstallationKey)
				{
					ret = RegQueryInfoKey(hKey2, NULL, NULL, NULL, NULL, NULL, NULL, &Values, NULL, NULL, NULL, NULL);
					if (ret != ERROR_SUCCESS)
						break;
					if (Values > 0)
						break;
				}
				// at this point, we presumably have a key that is empty of values, so it should be safe to delete
				if (RegDeleteKey(hKey2, RegKey) != ERROR_SUCCESS)
				{
					Err = GetLastError();
					temp.Format(IDS_REG_DELETE_KEY2, RegKey, Host);
					temp += LookupError(Err);
					ReportError(temp);
				}
				RegCloseKey(hKey2);
			}
		}
	}
	// end new
}

void CStrcmApp::StopServiceOperation(LPCTSTR CmdFile, LPCTSTR Host, CWnd* wnd)
{
	CString ServiceNameList = GetSetting(CmdFile, _T("Install"), _T("ServiceNames"));
	StopServices(ServiceNameList, Host, wnd);
}

CString CStrcmApp::StopServices(LPCTSTR ServiceList, LPCTSTR Host, CWnd* wnd)
{
	SC_HANDLE sch = NULL;
	SC_HANDLE sch2 = NULL;
	SERVICE_STATUS ssStatus;
	CString Services = ServiceList;
	CString ServiceName = "";
	CString temp;
	DWORD dwOldCheckPoint, dwStartTickCount, dwWaitTime, dwReturn;
	int i;
	BOOL FoundService = FALSE;
	DWORD Err;

	temp.Format(_T("Stopping remote control service on %s..."), Host);
	DisplayProgressMsg(wnd, temp);

	sch = OpenSCManager(Host, SERVICES_ACTIVE_DATABASE, SC_MANAGER_CONNECT);
	if (sch == NULL)
	{
		Err = GetLastError();
		temp.LoadString(IDS_OPEN_SERVICE_DB_ERROR);
		temp += LookupError(Err);
		ReportError(temp);
		goto bailout;
	}

	// append a delimiter to make parsing easier
	if (Services.Right(1) != ';')
		Services += ';';
	
	for (i=0; i < Services.GetLength(); i++)
	{
		if (Services[i] != ';')
			ServiceName += Services[i];
		else
		{
			sch2 = OpenService(sch, ServiceName, SERVICE_STOP | SERVICE_QUERY_STATUS);
			if (sch2 != NULL)
			{
				FoundService = TRUE;
				if (!ControlService(sch2, SERVICE_CONTROL_STOP, &ssStatus))
				{
					dwReturn = GetLastError();
					if (dwReturn != ERROR_SERVICE_NOT_ACTIVE)
					{
						Err = GetLastError();
						temp.LoadString(IDS_STOP_SERVICE_ERROR);
						temp += LookupError(Err);
						ReportError(temp);
						ServiceName = "";
					}
					goto bailout;
				}
				
				// we first check to see what the current status is.
				if (QueryServiceStatus(sch2, &ssStatus))
				{
					// we should wait around for a while to make sure its started
					dwStartTickCount = GetTickCount();
					dwOldCheckPoint = ssStatus.dwCheckPoint;
					while (ssStatus.dwCurrentState == SERVICE_STOP_PENDING) 
					{ 
						// Do not wait longer than the wait hint. A good interval is one tenth
						// the wait hint, but no less than 1 second and no more than 10 seconds. 
 
						dwWaitTime = ssStatus.dwWaitHint / 10;
						if (dwWaitTime < 1000)
							dwWaitTime = 1000;
						else if (dwWaitTime > 10000)
							dwWaitTime = 10000;
						Sleep(dwWaitTime);

						// Check the status again. 
 						if (!QueryServiceStatus(sch2, &ssStatus))
							break; 
 
						if (ssStatus.dwCheckPoint > dwOldCheckPoint)
						{
							// The service is making progress.
							dwStartTickCount = GetTickCount();
							dwOldCheckPoint = ssStatus.dwCheckPoint;
						}
						else
						{
							if (GetTickCount() - dwStartTickCount > ssStatus.dwWaitHint)
								// No progress made within the wait hint
								break;
						}
					}
				}

				// at this point, we are going to ASSume that the service stopped, but wait 
				// some more just to be sure, as some services don't report themselves very well (like VNC)
				Sleep(2000);

				// check one more time to make sure
				if (!QueryServiceStatus(sch2, &ssStatus))
				{
					Err = GetLastError();
					temp.LoadString(IDS_QUERY_SERVICE_STOP_ERROR);
					temp += LookupError(Err);
					ReportError(temp);
					ServiceName = "";
					dwReturn = GetLastError();
				}
				else
				{
					if (ssStatus.dwCurrentState != SERVICE_STOPPED)
					{
						temp.LoadString(IDS_STOP_VERIFY_ERROR);
						ReportError(temp);
						ServiceName = "";
					}
				}
				goto bailout;
			}
			else
			{
				dwReturn = GetLastError();
				if (dwReturn != ERROR_SERVICE_DOES_NOT_EXIST)
				{
					temp.LoadString(IDS_OPEN_SERVICE_ERROR);
					temp + LookupError(dwReturn);
					ReportError(temp);
				}
			}
			ServiceName = "";
		}
	}

bailout:
	if (sch)
		CloseServiceHandle(sch);
	if (sch2)
		CloseServiceHandle(sch2);

	if (ServiceName.GetLength() > 0)
	{
		temp.Format(_T("   ...Stopped %s."), ServiceName);
		DisplayProgressMsg(wnd, temp);
	}
	if (!FoundService)
	{
		temp.LoadString(IDS_NO_SERVICES_STOPPED);
		ReportError(temp);
	}
	return ServiceName;
}

void CStrcmApp::StartServiceOperation(LPCTSTR CmdFile, LPCTSTR Host, CWnd* wnd)
{
	CString ServiceNameList = GetSetting(CmdFile, _T("Install"), _T("ServiceNames"));
	StartServices(ServiceNameList, Host, wnd);
}

CString CStrcmApp::StartServices(LPCTSTR ServiceList, LPCTSTR Host, CWnd* wnd)
{
	SC_HANDLE sch = NULL;
	SC_HANDLE sch2 = NULL;
	SERVICE_STATUS ssStatus;
	CString Services = ServiceList;
	CString ServiceName = "";
	DWORD dwOldCheckPoint, dwStartTickCount, dwWaitTime, dwReturn;
	int i;

	CString temp;
	temp.Format(_T("Starting remote control service on %s..."), Host);
	DisplayProgressMsg(wnd, temp);
	sch = OpenSCManager(Host, SERVICES_ACTIVE_DATABASE, SC_MANAGER_CONNECT);
	if (sch == NULL)
	{
		// error opening service manager...we could report the error but what would we
		// do about it ?
		dwReturn = GetLastError();
		goto bailout;
	}

	// append a delimiter to make parsing easier
	if (Services.Right(1) != ';')
		Services += ';';
	
	for (i=0; i < Services.GetLength(); i++)
	{
		if (Services[i] != ';')
			ServiceName += Services[i];
		else
		{
			sch2 = OpenService(sch, ServiceName, SERVICE_START | SERVICE_QUERY_STATUS);
			if (sch2 != NULL)
			{
				// we might need to return an error if we can't start it, but maybe its 
				// already started...
				if (!StartService(sch2, 0, NULL))
					goto bailout;

				// we first check to see what the current status is.
				if (!QueryServiceStatus(sch2, &ssStatus))
					goto bailout;

				// we should wait around for a while to make sure its started
				dwStartTickCount = GetTickCount();
				dwOldCheckPoint = ssStatus.dwCheckPoint;
				while (ssStatus.dwCurrentState == SERVICE_START_PENDING) 
				{ 
					// Do not wait longer than the wait hint. A good interval is one tenth
					// the wait hint, but no less than 1 second and no more than 10 seconds. 
 
					dwWaitTime = ssStatus.dwWaitHint / 10;
					if (dwWaitTime < 1000)
						dwWaitTime = 1000;
					else if (dwWaitTime > 10000)
						dwWaitTime = 10000;
					Sleep(dwWaitTime);

					// Check the status again. 
 			        if (!QueryServiceStatus(sch2, &ssStatus))
						break; 
 
			        if (ssStatus.dwCheckPoint > dwOldCheckPoint)
					{
						// The service is making progress.
						dwStartTickCount = GetTickCount();
						dwOldCheckPoint = ssStatus.dwCheckPoint;
					}
					else
					{
						if (GetTickCount() - dwStartTickCount > ssStatus.dwWaitHint)
							// No progress made within the wait hint
							break;
					}
				}
				// at this point, we are going to ASSume that the service started, but wait 
				// some more just to be sure, as some services don't report themselves very well (like VNC)
				Sleep(2000);
				goto bailout;
			}
			else
			{
				// we could not open the service, but we can't really do anything about it
				// as it might not exist, or maybe not enough rights
				dwReturn = GetLastError();
			}
			ServiceName = "";
		}
	}

bailout:
	if (sch)
		CloseServiceHandle(sch);
	if (sch2)
		CloseServiceHandle(sch2);
	if (ServiceName.GetLength() > 0)
	{
		CString temp;
		temp.Format(_T("   ...Started %s."), ServiceName);
		DisplayProgressMsg(wnd, temp);
	}
	return ServiceName;
}


void CStrcmApp::SetVncOptions(LPCTSTR CmdFile, LPCTSTR Host, LPCTSTR Pwd, BOOL ShowDialog)
{
	CVNCOptions Options;
	CString Password, ePwd, temp, Setting, Value, SpecialKey;
	DWORD Err, BigBlobSize, dwValue;
	HKEY hKey = NULL;
	HKEY hKey2 = NULL;
	HKEY hKeySpecial = NULL;
	LPTSTR BigBlob = NULL;
	int i, j, f;
	TCHAR* stopstring;
	BOOL FoundSpecial;
	BOOL ForceX64;

	ForceX64 = GetForceX64Setting(CmdFile);

	Password = Pwd;
	
	if (ShowDialog)
	{
		Options.m_PollConsoleOnly = GetSettingInt(CmdFile, _T("VNC Settings"), _T("OnlyPollConsole"), 1);
		Options.m_PollEventOnly = GetSettingInt(CmdFile, _T("VNC Settings"), _T("OnlyPollOnEvent"), 0);
		Options.m_PollForeground = GetSettingInt(CmdFile, _T("VNC Settings"), _T("PollForeground"), 1);
		Options.m_PollFullScreen = GetSettingInt(CmdFile, _T("VNC Settings"), _T("PollFullScreen"), 0);
		Options.m_PollUnderCursor = GetSettingInt(CmdFile, _T("VNC Settings"), _T("PollUnderCursor"), 0);
		Options.m_SocketConnect = GetSettingInt(CmdFile, _T("VNC Settings"), _T("SocketConnect"), 1);
		Options.m_PortNumber = GetSettingInt(CmdFile, _T("VNC Settings"), _T("PortNumber"), 5900);
		if (Options.m_PortNumber >= VNC_PORT_OFFSET)
			Options.m_PortNumber = Options.m_PortNumber - VNC_PORT_OFFSET;
		Options.m_AutoPortSelect = GetSettingInt(CmdFile, _T("VNC Settings"), _T("AutoPortSelect"), 1);
		Options.m_DisableRemoteKP = !(GetSettingInt(CmdFile, _T("VNC Settings"), _T("InputsEnabled"), 1));
		Options.m_DisableLocalKP = GetSettingInt(CmdFile, _T("VNC Settings"), _T("LocalInputsDisabled"), 0);
		if (Options.DoModal() != IDOK)
			return;
		if (Options.m_Pwd.GetLength() > 0)
			Password = Options.m_Pwd;
		// now we save the options back into the file.  Why ?  That way, we can use the code
		// below to read all of the settings from the file, even the ones that are entered manually
		SetSettingInt(CmdFile, _T("VNC Settings"), _T("OnlyPollConsole"), Options.m_PollConsoleOnly);
		SetSettingInt(CmdFile, _T("VNC Settings"), _T("OnlyPollOnEvent"), Options.m_PollEventOnly);
		SetSettingInt(CmdFile, _T("VNC Settings"), _T("PollForeground"), Options.m_PollForeground);
		SetSettingInt(CmdFile, _T("VNC Settings"), _T("PollFullScreen"), Options.m_PollFullScreen);
		SetSettingInt(CmdFile, _T("VNC Settings"), _T("PollUnderCursor"), Options.m_PollUnderCursor);
		SetSettingInt(CmdFile, _T("VNC Settings"), _T("SocketConnect"), Options.m_SocketConnect);
		Options.m_PortNumber += VNC_PORT_OFFSET;
		SetSettingInt(CmdFile, _T("VNC Settings"), _T("PortNumber"), Options.m_PortNumber);
		SetSettingInt(CmdFile, _T("VNC Settings"), _T("AutoPortSelect"), Options.m_AutoPortSelect);
		SetSettingInt(CmdFile, _T("VNC Settings"), _T("InputsEnabled"), !(Options.m_DisableRemoteKP));
		SetSettingInt(CmdFile, _T("VNC Settings"), _T("LocalInputsDisabled"), Options.m_DisableLocalKP);
	}

	if (Err = RegConnectRegistry(Host, HKEY_LOCAL_MACHINE, &hKey) != ERROR_SUCCESS)
	{
		temp.Format(IDS_CONNECT_REGISTRY_ERROR, Host);
		temp += LookupError(Err);
		ReportError(temp);
		goto bailout;
	}

	SpecialKey = GetSetting(CmdFile, _T("Install"), _T("RegistryKeyLocation"));
	if (SpecialKey.GetLength() == 0)
		SpecialKey = "SOFTWARE\\ORL\\WinVNC3\\Default";

	if (Err = RegCreateKeyEx(hKey, SpecialKey, 0, NULL, 0, GetRegAccessRights(KEY_SET_VALUE, ForceX64), NULL, &hKey2, NULL) != ERROR_SUCCESS) 
	{
		temp.Format(IDS_OPEN_REGISTRY_KEY_ERROR, SpecialKey);
		temp += LookupError(Err);
		ReportError(temp);
		goto bailout;
	}
	
	if (Password.GetLength() > 0)
	{
		ePwd = EncryptVncPassword(Password);
		Setting = "Password";
		if (Err = RegSetValueEx(hKey2, Setting, 0, REG_BINARY, (CONST BYTE*) (LPCTSTR) ePwd, MAX_VNC_PWD_LEN) != ERROR_SUCCESS)
		{
			temp.Format(IDS_SET_REGISTRY_KEY_ERROR, Setting);
			temp += LookupError(Err);
			ReportError(temp);
		}
	}

	// now we do something that may seem a bit strange.  Rather than reading hard-coded
	// setting from the .RCM file, we instead read all of the settings, and write them all 
	// out.  Why ?  VNC uses a number of documented settings that cannot be set in the GUI
	// and this way, anyone can add any registry value that they want to into the .RCM file
	// and the GUI of the STRCM program does not have to be changed.
	
	BigBlob = (LPTSTR) HeapAlloc(GetProcessHeap(), 0, BIG_BLOB_SIZE); //more than enough room
	if (BigBlob == NULL)
	{
		// TODO: display error...should never happen
		goto bailout;
	}

	BigBlobSize = GetPrivateProfileSection(_T("VNC Settings"), BigBlob, BIG_BLOB_SIZE, CmdFile);

	// NOTES: 
	//		By convention, I rarely use the continue statement, but the code started 
	//		drifting too far to the right so it was used below to keep things more cleaner
	//
	//		I rarely use the RegFlushKey command, but thought that it would be safer to use
	//		it if the keys are removed later in an uninstall.  I don't know if RegFlushKey
	//		does anything at all

	// now go through our buffer of all of the settings, and write each value
	Setting = "";
	for (i=0; i < BigBlobSize; i++)
	{
		if (BigBlob[i] != '\0')
		{
			Setting += BigBlob[i];
			continue;
		}
		
		// look to see if the default location has been overridden
		FoundSpecial = FALSE;
		f = Setting.ReverseFind('\\');
		if (f != -1)
		{
			// dang, now we have to do more work
			FoundSpecial = TRUE;

			SpecialKey = Setting.Left(f);
			Setting = Setting.Mid(f + 1);
			
			if (Err = RegCreateKeyEx(hKey, SpecialKey, 0, NULL, 0, GetRegAccessRights(KEY_SET_VALUE, ForceX64), NULL, &hKeySpecial, NULL) != ERROR_SUCCESS) 
			{
				temp.Format(IDS_OPEN_REGISTRY_KEY_ERROR, SpecialKey);
				temp += LookupError(Err);
				ReportError(temp);
				goto bailout;
			}
		}

		// look for the '=' and parse out the settings vs. the value
		f = Setting.Find('=');
		if (f == -1)
			continue;
		
		Value = Setting.Mid(f+1);
		Setting = Setting.Left(f);
		if (Setting.CompareNoCase(_T("password")) == 0)
		{
			// here we have to convert the character hex password characters
			// into binary then store...yuck.
			temp = "";
			ePwd = "";
			for (j=0; j < Value.GetLength(); j++)
			{
				if (Value[j] != ' ')  //skip any blank spaces
				{
					temp += Value[j];
					if (temp.GetLength() == 2) // if we have a 2-char hex char
					{
						// convert the hex char to a number
						dwValue = _tcstoul(temp, &stopstring, 16);
						ePwd += (char) dwValue; // v1.14
						temp = "";
					}
				}
			}
			// now set the password
			if (Err = RegSetValueEx((FoundSpecial) ? hKeySpecial : hKey2, Setting, 0, REG_BINARY, (CONST BYTE*) (LPCTSTR) ePwd, MAX_VNC_PWD_LEN) != ERROR_SUCCESS)
			{
				temp.Format(IDS_SET_REGISTRY_KEY_ERROR, Setting);
				temp += LookupError(Err);
				ReportError(temp);
			}
		}
		else if (Setting.CompareNoCase(_T("authhosts")) == 0) //special case for REG_SZ types
		{
			if (Err = RegSetValueEx((FoundSpecial) ? hKeySpecial : hKey2, Setting, 0, REG_SZ, (CONST BYTE*) (LPCTSTR) Value, Value.GetLength() + 1) != ERROR_SUCCESS)
			{
				temp.Format(IDS_SET_REGISTRY_KEY_ERROR, Setting);
				temp += LookupError(Err);
				ReportError(temp);
			}
		}
		//REG_BINARY...new in v1.13
		else if (Value.Find(_T("REG_BINARY:")) >= 0)
		{
			//Get rid of the reg_binary: portion
			Value.Replace("REG_BINARY:", NULL);

			// here we have to convert the character hex password characters
			// into binary then store...yuck.
			temp = "";
			ePwd = "";
			for (j=0; j < Value.GetLength(); j++)
			{
				if (Value[j] != ' ')  //skip any blank spaces
				{
					temp += Value[j];
					if (temp.GetLength() == 2) // if we have a 2-char hex char
					{
						// convert the hex char to a number
						dwValue = _tcstoul(temp, &stopstring, 16);
						ePwd += (char) dwValue; // v1.14
						temp = "";
					}
				}
			}
			// now set the value
			if (Err = RegSetValueEx((FoundSpecial) ? hKeySpecial : hKey2, Setting, 0, REG_BINARY, (CONST BYTE*) (LPCTSTR) ePwd, ePwd.GetLength()) != ERROR_SUCCESS)
			{
				temp.Format(IDS_SET_REGISTRY_KEY_ERROR, Setting);
				temp += LookupError(Err);
				ReportError(temp);
			}
		}
		else if (Value.Find(_T("REG_SZ:")) >= 0) //special case for REG_SZ types...new in v1.13
		{
			//Get rid of the reg_sz: portion
			Value.Replace("REG_SZ:", NULL);

			if (Err = RegSetValueEx((FoundSpecial) ? hKeySpecial : hKey2, Setting, 0, REG_SZ, (CONST BYTE*) (LPCTSTR) Value, Value.GetLength() + 1) != ERROR_SUCCESS)
			{
				temp.Format(IDS_SET_REGISTRY_KEY_ERROR, Setting);
				temp += LookupError(Err);
				ReportError(temp);
			}
		}
		else // all others assume REG_DWORD
		{
			dwValue = _ttol(Value);
			if (Err = RegSetValueEx((FoundSpecial) ? hKeySpecial : hKey2, Setting, 0, REG_DWORD, (CONST BYTE*) &dwValue, sizeof(DWORD)) != ERROR_SUCCESS)
			{
				temp.Format(IDS_SET_REGISTRY_KEY_ERROR, Setting);
				temp += LookupError(Err);
				ReportError(temp);
			}
		}
		Setting = "";

		// close our special handle if we had one
		if (hKeySpecial)
		{
			RegFlushKey(hKeySpecial);
			RegCloseKey(hKeySpecial);
			hKeySpecial = NULL;
		}
	}

bailout:
	if (BigBlob)
		HeapFree(GetProcessHeap(), 0, BigBlob);
	if (hKey)
	{
		RegFlushKey(hKey);
		RegCloseKey(hKey);
	}
	if (hKey2)
	{
		RegFlushKey(hKey2);
		RegCloseKey(hKey2);
	}
}

void CStrcmApp::SetOptions(LPCTSTR CmdFile, LPCTSTR Host, CWnd* wnd)
{
	CString SoftwareType = GetSetting(CmdFile, _T("General"), _T("SoftwareType"));
	CString temp;
		
	// here we set the options depending upon the type of software
	if (SoftwareType.CompareNoCase(_T("VNC")) == 0)
		SetVncOptions(CmdFile, Host, _T(""), TRUE);
	temp.Format(_T("Option configuration complete for %s."), Host);
	DisplayProgressMsg(wnd, temp);
}

void CStrcmApp::DisplayProgressMsg(CWnd* wnd, LPCTSTR Msg)
{
	if (wnd)
	{
		int i = ((CListBox*) wnd)->AddString(Msg);
		((CListBox*) wnd)->UpdateWindow();
		((CListBox*) wnd)->SetCurSel(i);
	}
	else
	{
		// other logging options can go here
	}
}